from .client import NetztransparenzClient as NetztransparenzClient
from .vermarktung_client import VermarktungClient as VermarktungClient
from .hochrechnung_client import HochrechnungClient as HochrechnungClient
from .dienstleistungen_client import DienstleistungenClient as DienstleistungenClient
from .nrvsaldo_client import NrvSaldoClient as NrvSaldoClient



"""
Client for all Endpoints of the "NrvSaldo/" Group
"""

from netztransparenz.base_client import BaseNtClient
import requests
import datetime as dt
import io

import pandas as pd

_nrvsaldo_date_format = "%d.%m.%Y %H:%M %Z"


class NrvSaldoClient(BaseNtClient):
    def __init__(self, client_id, client_pass):
        super().__init__(client_id, client_pass)

    def _basic_read_nrvsaldo(
        self,
        resource_url,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Internal method to read data in the format of most /nrvsaldo dataseries.
        Target format is: Dates separated in "Datum", "von", "bis", "Zeitzone".
        Return a pandas Dataframe with data of the endpoint specified with resource_url.
        If either dt_begin or dt_end is None, all available data will be queried.

            resource_url -- url of the endpoint without the base url and without leading or trailing "/"
            earliest_data -- first datapoint in the source collection
            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/data/{resource_url}"
        if (dt_begin is not None) and (dt_end is not None):
            start_of_data = dt_begin.strftime(self._api_date_format)
            end_of_data = dt_end.strftime(self._api_date_format)
            url = url + f"/{start_of_data}/{end_of_data}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A.", "N.E.", ""],
        )

        if transform_dates:
            df["von"] = pd.to_datetime(
                df["Datum"] + " " + df["von"] + " " + df["Zeitzone"],
                format=_nrvsaldo_date_format,
            ).dt.tz_convert(None)
            df["bis"] = pd.to_datetime(
                df["Datum"] + " " + df["bis"] + " " + df["Zeitzone"],
                format=_nrvsaldo_date_format,
            ).dt.tz_convert(None)
            # The end of timeframes may be 00:00 of the next day witch is not correctly represented in timestamps
            df["bis"] = df["bis"].where(
                df["bis"].dt.time != dt.time(0, 0), df["bis"] + dt.timedelta(days=1)
            )
            df = df.drop(["Datum", "Zeitzone"], axis=1).set_index("von")
        return df

    def traffic_light(self, dt_begin: dt.datetime, dt_end: dt.datetime):
        """
        Return a pandas Dataframe with data of the endpoint /TrafficLight.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
        """
        url = f"{self._API_BASE_URL}/data/TrafficLight/{dt_begin.strftime(self._api_date_format)}/{dt_end.strftime(self._api_date_format)}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_json(io.StringIO(response.text))

        return df

    def nrvsaldo_nrvsaldo_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/NRVSaldo/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/NRVSaldo/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_nrvsaldo_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/NRVSaldo/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/NRVSaldo/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_rzsaldo_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/RZSaldo/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/RZSaldo/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_rzsaldo_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/RZSaldo/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/RZSaldo/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_prl_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/PRL/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/PRL/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_prl_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/PRL/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/PRL/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_aktivierte_srl_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /Nrvsaldo/AktivierteSRL/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "Nrvsaldo/AktivierteSRL/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_aktivierte_srl_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /Nrvsaldo/AktivierteSRL/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "Nrvsaldo/AktivierteSRL/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_aktivierte_mrl_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/AktivierteMRL/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/AktivierteMRL/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_aktivierte_mrl_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/AktivierteMRL/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/AktivierteMRL/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_srl_optimierung_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/SRLOptimierung/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/SRLOptimierung/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_srl_optimierung_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/SRLOptimierung/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/SRLOptimierung/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_mrl_optimierung_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/MRLOptimierung/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/MRLOptimierung/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_mrl_optimierung_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/MRLOptimierung/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/MRLOptimierung/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_mrl_mol_abweichungen_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/MrlMolAbweichungen/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/MrlMolAbweichungen/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_srl_mol_abweichungen_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/SrlMolAbweichungen/betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/data/NrvSaldo/SrlMolAbweichungen/betrieblich"
        if (dt_begin is not None) and (dt_end is not None):
            start_of_data = dt_begin.strftime(self._api_date_format)
            end_of_data = dt_end.strftime(self._api_date_format)
            url = url + f"/{start_of_data}/{end_of_data}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A.", ""],
        )

        if transform_dates:
            df["von"] = pd.to_datetime(
                df["Datum von"] + " " + df["Uhrzeit von"] + " " + df["Zeitzone von"],
                format=_nrvsaldo_date_format,
            ).dt.tz_convert(None)
            df["bis"] = pd.to_datetime(
                df["Datum bis"] + " " + df["Uhrzeit bis"] + " " + df["Zeitzone bis"],
                format=_nrvsaldo_date_format,
            ).dt.tz_convert(None)
            # The end of timeframes may be 00:00 of the next day witch is not correctly represented in timestamps
            df["bis"] = df["bis"].where(
                df["bis"].dt.time != dt.time(0, 0), df["bis"] + dt.timedelta(days=1)
            )
            df = df.drop(
                [
                    "Datum von",
                    "Uhrzeit von",
                    "Zeitzone von",
                    "Datum bis",
                    "Uhrzeit bis",
                    "Zeitzone bis",
                ],
                axis=1,
            ).set_index("von")
        return df

    def nrvsaldo_difference_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/Difference/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/Difference/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_difference_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/Difference/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/Difference/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_abschaltbare_lasten_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/AbschaltbareLasten/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/AbschaltbareLasten/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_abschaltbare_lasten_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/AbschaltbareLasten/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/AbschaltbareLasten/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_zusatzmassnahmen_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/Zusatzmassnahmen/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/Zusatzmassnahmen/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_zusatzmassnahmen_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/Zusatzmassnahmen/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/Zusatzmassnahmen/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_nothilfe_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/Nothilfe/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/Nothilfe/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_nothilfe_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/Nothilfe/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/Nothilfe/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_rebap_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/reBAP/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/reBAP/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_aep_module_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/AEPModule/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/AEPModule/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_aep_schaetzer_betrieblich(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/AepSchaetzer/Betrieblich/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/AepSchaetzer/Betrieblich", dt_begin, dt_end, transform_dates
        )

    def nrvsaldo_finanzielle_wirkung_aep_module_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/FinanzielleWirkungAEPModule/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/FinanzielleWirkungAEPModule/Qualitaetsgesichert",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def nrvsaldo_voaa_qualitaetsgesichert(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NrvSaldo/VoAA/Qualitaetsgesichert/.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nrvsaldo(
            "NrvSaldo/VoAA/Qualitaetsgesichert", dt_begin, dt_end, transform_dates
        )



"""
Client for all /vermarktung/ Endpoints and other Endpoints with market data like /Spotmarkpreise
"""

from netztransparenz.base_client import BaseNtClient
import requests
import datetime as dt
import io

import pandas as pd

_csv_date_format = "%Y-%m-%d %H:%M %Z"


class VermarktungClient(BaseNtClient):
    def __init__(self, client_id, client_pass):
        super().__init__(client_id, client_pass)

    def _basic_read_vermarktung(
        self,
        resource_url,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
        dateformat=_csv_date_format,
    ):
        """
        Internal method to read data in the format of most /vermarktung dataseries.
        Target format is: Dates separated in "datum", "von", "bis", "zeitzone".
        Return a pandas Dataframe with data of the endpoint specified with resource_url.
        If either dt_begin or dt_end is None, all available data will be queried.

            resource_url -- url of the endpoint without the base url and without leading or trailing "/"
            earliest_data -- first datapoint in the source collection
            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/{resource_url}"
        if (dt_begin is not None) and (dt_end is not None):
            start_of_data = dt_begin.strftime(self._api_date_format)
            end_of_data = dt_end.strftime(self._api_date_format)
            url = f"{self._API_BASE_URL}/{resource_url}/{start_of_data}/{end_of_data}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A."],
        )

        if transform_dates:
            df["von"] = pd.to_datetime(
                df["Datum"] + " " + df["von"] + " " + df["Zeitzone von"],
                format=dateformat,
                utc=True,
            )
            df["bis"] = pd.to_datetime(
                df["Datum"] + " " + df["bis"] + " " + df["Zeitzone bis"],
                format=dateformat,
                utc=True,
            )
            # The end of timeframes may be 00:00 of the next day which is not correctly represented in timestamps
            df["bis"] = df["bis"].where(
                df["bis"].dt.time != dt.time(0, 0), df["bis"] + dt.timedelta(days=1)
            )
            df = df.drop(["Datum", "Zeitzone von", "Zeitzone bis"], axis=1).set_index(
                "von"
            )

        return df

    def _basic_read_negative_preise(
        self,
        resource_url,
        dt_begin: dt.datetime,
        dt_end: dt.datetime,
        transform_dates=False,
    ):
        """
        Internal method to read data in the format of the /NegativePreise dataseries.
        Target format is: Only two columns "Datum" and "Negativ", start and end dates not optional.
        Return a pandas Dataframe with data of the endpoint specified with resource_url.

            resource_url -- url of the endpoint without the base url and without leading or trailing "/"
            earliest_data -- first datapoint in the source collection
            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times as a string in the format "%Y-%m-%d %H:%M"
                               if this option resolves to "True" the times will be transformed into a
                               fully qualified timestamp. (default: False)
        """
        url = f"{self._API_BASE_URL}/{resource_url}/{dt_begin.strftime(self._api_date_format)}/{dt_end.strftime(self._api_date_format)}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A."],
        )

        if transform_dates:
            df["Datum"] = pd.to_datetime(df["Datum"], format="%Y-%m-%d %H:%M")

        return df

    def vermarktung_differenz_einspeiseprognose(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/DifferenzEinspeiseprognose.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/DifferenzEinspeiseprognose",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def vermarktung_inanspruchnahme_ausgleichsenergie(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/InanspruchnahmeAusgleichsenergie.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/InanspruchnahmeAusgleichsenergie",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def vermarktung_untertaegige_strommengen(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/UntertaegigeStrommengen.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/UntertaegigeStrommengen",
            dt_begin,
            dt_end,
            transform_dates,
        )

    def vermarktung_epex(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/VermarktungEpex.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-12-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/VermarktungEpex", dt_begin, dt_end, transform_dates
        )

    def vermarktung_exaa(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/VermarktungExaa.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-12-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/VermarktungExaa", dt_begin, dt_end, transform_dates
        )

    def vermarktung_solar(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/VermarktungsSolar.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2013-12-31T23:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/VermarktungsSolar", dt_begin, dt_end, transform_dates
        )

    def vermarktung_wind(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/VermarktungsWind.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2013-12-31T23:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/VermarktungsWind", dt_begin, dt_end, transform_dates
        )

    def vermarktung_sonstige(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /vermarktung/VermarktungsSonstige.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2013-12-31T23:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/vermarktung/VermarktungsSonstige", dt_begin, dt_end, transform_dates
        )

    def spotmarktpreise(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /Spotmarktpreise.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2013-12-31T23:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_vermarktung(
            "data/Spotmarktpreise",
            dt_begin,
            dt_end,
            transform_dates=transform_dates,
            dateformat="%d.%m.%Y %H:%M %Z",
        )

    def negative_preise_1h(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NegativePreise/1.
        dt_begin and dt_end should not be too far apart since the api may throw an error for spans more than a year.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times as a string in the format "%Y-%m-%d %H:%M"
                               if this option resolves to "True" the times will be transformed into a
                               fully qualified timestamp. (default: False)
        """
        return self._basic_read_negative_preise(
            "data/NegativePreise/1", dt_begin, dt_end, transform_dates
        )

    def negative_preise_3h(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NegativePreise/1.
        dt_begin and dt_end should not be too far apart since the api may throw an error for spans more than a year.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times as a string in the format "%Y-%m-%d %H:%M"
                               if this option resolves to "True" the times will be transformed into a
                               fully qualified timestamp. (default: False)
        """
        return self._basic_read_negative_preise(
            "data/NegativePreise/3", dt_begin, dt_end, transform_dates
        )

    def negative_preise_4h(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NegativePreise/1.
        dt_begin and dt_end should not be too far apart since the api may throw an error for spans more than a year.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times as a string in the format "%Y-%m-%d %H:%M"
                               if this option resolves to "True" the times will be transformed into a
                               fully qualified timestamp. (default: False)
        """
        return self._basic_read_negative_preise(
            "data/NegativePreise/4", dt_begin, dt_end, transform_dates
        )

    def negative_preise_6h(
        self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False
    ):
        """
        Return a pandas Dataframe with data of the endpoint /NegativePreise/1.
        dt_begin and dt_end should not be too far apart since the api may throw an error for spans more than a year.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times as a string in the format "%Y-%m-%d %H:%M"
                               if this option resolves to "True" the times will be transformed into a
                               fully qualified timestamp. (default: False)
        """
        return self._basic_read_negative_preise(
            "data/NegativePreise/6", dt_begin, dt_end, transform_dates
        )

    def jahresmarktpraemie(self, year: int | None = None):
        """
        Return a pandas Dataframe with data of the endpoint /Jahresmarktpraemie.
        If no year is given, all available data will be queried.

            year -- int representation of the year to get the data for (earliest data: 2020)
        """
        url = f"{self._API_BASE_URL}/data/Jahresmarktpraemie/"
        if year is not None:
            url = url + str(year)

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A."],
        )
        return df

    def marktpraemie(self, dt_begin: dt.date, dt_end: dt.date, transform_dates=False):
        """
        Return a pandas Dataframe with data of the endpoint /marktpraemie.
        Year & month of both start and end are extracted from dt_begin and dt_end.

            dt_begin -- date object for start of data (day will be ignored)
            dt_end -- date object for end of data (day will be ignored)
            transform_dates -- data contains months in th format "1/2012"
        """
        url = f"{self._API_BASE_URL}/data/marktpraemie/{dt_begin.month}/{dt_begin.year}/{dt_end.month}/{dt_end.year}"
        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A.", ""],
        )
        return df

    def id_aep(self, dt_begin: dt.datetime, dt_end: dt.datetime, transform_dates=False):
        """
        Return a pandas Dataframe with data of the endpoint /IdAep.
        The raw data contains the columns: "Datum von;(Uhrzeit) von;Zeitzone;(Uhrzeit) bis;Zeitzone;ID AEP in €/MWh"
        Since pandas can not handle ambigious columns the 'Zeitzone' columns are renamed to 'Zeitzone von' and 'Zeitzone bis' respectively.

            dt_begin -- datetime object for start of data in UTC
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/data/IdAep/{dt_begin.strftime(self._api_date_format)}/{dt_end.strftime(self._api_date_format)}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A."],
        )
        df.rename(
            columns={"Zeitzone": "Zeitzone von", "Zeitzone.1": "Zeitzone bis"},
            inplace=True,
        )

        if transform_dates:
            df["von"] = pd.to_datetime(
                df["Datum von"] + " " + df["(Uhrzeit) von"] + " " + df["Zeitzone von"],
                format=_csv_date_format,
                utc=True,
            )
            df["bis"] = pd.to_datetime(
                df["Datum von"] + " " + df["(Uhrzeit) bis"] + " " + df["Zeitzone bis"],
                format=_csv_date_format,
                utc=True,
            )
            # The end of timeframes may be 00:00 of the next day which is not correctly represented in timestamps
            df["bis"] = df["bis"].where(
                df["bis"].dt.time != dt.time(0, 0), df["bis"] + dt.timedelta(days=1)
            )
            df = df.drop(
                ["Datum von", "Zeitzone von", "Zeitzone bis"], axis=1
            ).set_index("von")

        return df



"""
Base class for all other clients. Handles login and health check.
"""

import requests
import logging

log = logging.getLogger("BaseNtClient")
_ACCESS_TOKEN_URL = "https://identity.netztransparenz.de/users/connect/token"


class BaseNtClient:
    def __init__(self, client_id, client_pass):
        self._API_BASE_URL = "https://ds.netztransparenz.de/api/v1"
        self._api_date_format = "%Y-%m-%dT%H:%M:%S"
        self._csv_date_format = "%Y-%m-%d %H:%M %Z"

        response = requests.post(
            _ACCESS_TOKEN_URL,
            data={
                "grant_type": "client_credentials",
                "client_id": client_id,
                "client_secret": client_pass,
            },
        )

        if response.ok:
            self.token = response.json()["access_token"]
        else:
            message = (
                f"Error retrieving token\n{response.status_code}:{response.reason}"
            )
            log.error(message)
            raise Exception(f"Login failed. {message}")

    def check_health(self):
        """
        Return the text response of the API health endpoint.
        Any Response but "OK" indicates a problem.
        """

        url = f"{self._API_BASE_URL}/health"
        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        return response.text



"""
Inofficial wrapper for the API of the www.netztransparenz.de platform.

To access the API it is required to set up a free account and client in the
Netztransparenz extranet. (see: https://www.netztransparenz.de/en/Web-API)
"""

from .vermarktung_client import VermarktungClient
from .hochrechnung_client import HochrechnungClient
from .dienstleistungen_client import DienstleistungenClient
from .nrvsaldo_client import NrvSaldoClient


class NetztransparenzClient(
    VermarktungClient, HochrechnungClient, DienstleistungenClient, NrvSaldoClient
):
    def __init__(self, client_id, client_pass):
        super().__init__(client_id, client_pass)



"""
Client for all /<online>hochrechnung/ and /prognose/ Endpoints
"""

from netztransparenz.base_client import BaseNtClient
import requests
import datetime as dt
import io

import pandas as pd

_csv_date_format = "%Y-%m-%d %H:%M %Z"


class HochrechnungClient(BaseNtClient):
    def __init__(self, client_id, client_pass):
        super().__init__(client_id, client_pass)

    def _basic_read_nt(
        self,
        resource_url,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Internal method to read data in one of the common formats of th nt portal.
        Target format is: Dates separated in "Datum", "von", "Zeitzone von", "bis", "Zeitzone bis".
        Return a pandas Dataframe with data of the endpoint specified with resource_url.
        If either dt_begin or dt_end is None, all available data will be queried.

            resource_url -- url of the endpoint without the base url and without leading or trailing "/"
            earliest_data -- first datapoint in the source collection
            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/{resource_url}"
        if (dt_begin is not None) and (dt_end is not None):
            start_of_data = dt_begin.strftime(self._api_date_format)
            end_of_data = dt_end.strftime(self._api_date_format)
            url = f"{self._API_BASE_URL}/{resource_url}/{start_of_data}/{end_of_data}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            thousands=".",
            na_values=["N.A.", "N.E."],
        )

        if transform_dates:
            df["von"] = pd.to_datetime(
                df["Datum"] + " " + df["von"] + " " + df["Zeitzone von"],
                format=_csv_date_format,
                utc=True,
            )
            df["bis"] = pd.to_datetime(
                df["Datum"] + " " + df["bis"] + " " + df["Zeitzone bis"],
                format=_csv_date_format,
                utc=True,
            )
            # The end of timeframes may be 00:00 of the next day which is not correctly represented in timestamps
            df["bis"] = df["bis"].where(
                df["bis"].dt.time != dt.time(0, 0), df["bis"] + dt.timedelta(days=1)
            )
            df = df.drop(["Datum", "Zeitzone von", "Zeitzone bis"], axis=1).set_index(
                "von"
            )

        return df

    def hochrechnung_solar(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /hochrechnung/Solar.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/hochrechnung/Solar", dt_begin, dt_end, transform_dates
        )

    def hochrechnung_wind(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /hochrechnung/Wind.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/hochrechnung/Wind", dt_begin, dt_end, transform_dates
        )

    def online_hochrechnung_windonshore(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /OnlineHochrechnung/Windonshore.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/OnlineHochrechnung/Windonshore", dt_begin, dt_end, transform_dates
        )

    def online_hochrechnung_windoffshore(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /OnlineHochrechnung/Windoffshore.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/OnlineHochrechnung/Windoffshore", dt_begin, dt_end, transform_dates
        )

    def online_hochrechnung_solar(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /OnlineHochrechnung/Solar.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/OnlineHochrechnung/Solar", dt_begin, dt_end, transform_dates
        )

    def prognose_solar(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /prognose/Solar.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC (no values after: 2022-12-15T00:00:00)
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/prognose/Solar", dt_begin, dt_end, transform_dates
        )

    def prognose_wind(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /prognose/Wind.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC (no values after: 2022-12-15T00:00:00)
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "von" and "bis" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_nt(
            "data/prognose/Wind", dt_begin, dt_end, transform_dates
        )



"""
Client for all Endpoints of the "Systemdienstleistungen" Group
"""

from netztransparenz.base_client import BaseNtClient
import requests
import datetime as dt
import io

import pandas as pd


class DienstleistungenClient(BaseNtClient):
    def __init__(self, client_id, client_pass):
        super().__init__(client_id, client_pass)

    def _basic_read_systemdienstleistungen(
        self,
        resource_url,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Internal method to read data in the format of most 'systemdienstleistungen' dataseries.
        Target format is: Dates separated in "BEGINN_DATUM", "BEGINN_UHRZEIT", "ENDE_DATUM",
        "ENDE_UHRZEIT", "ZEITZONE_VON", "ZEITZONE_BIS".
        Return a pandas Dataframe with data of the endpoint specified with resource_url.
        If either dt_begin or dt_end is None, all available data will be queried.

            resource_url -- url of the endpoint without the base url and without leading or trailing "/"
            earliest_data -- first datapoint in the source collection
            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/data/{resource_url}"
        if (dt_begin is not None) and (dt_end is not None):
            start_of_data = dt_begin.strftime(self._api_date_format)
            end_of_data = dt_end.strftime(self._api_date_format)
            url = f"{self._API_BASE_URL}/data/{resource_url}/{start_of_data}/{end_of_data}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A."],
        )

        if transform_dates:
            df["BEGINN"] = pd.to_datetime(
                df["BEGINN_DATUM"]
                + " "
                + df["BEGINN_UHRZEIT"]
                + " "
                + df["ZEITZONE_VON"],
                format="%d.%m.%Y %H:%M %Z",
                utc=True,
            ).dt.tz_localize(None)
            df["ENDE"] = pd.to_datetime(
                df["ENDE_DATUM"] + " " + df["ENDE_UHRZEIT"] + " " + df["ZEITZONE_BIS"],
                format="%d.%m.%Y %H:%M %Z",
                utc=True,
            ).dt.tz_localize(None)
            df = df.drop(
                [
                    "BEGINN_DATUM",
                    "BEGINN_UHRZEIT",
                    "ENDE_DATUM",
                    "ENDE_UHRZEIT",
                    "ZEITZONE_VON",
                    "ZEITZONE_BIS",
                ],
                axis=1,
            )

        return df

    def _basic_read_abregelung(
        self,
        resource_url,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Internal method to read data in the format of most dataseries concerning power reduction.
        Target format is: Dates separated in "Datum", "Zeitzone", "von", "bis".
        Return a pandas Dataframe with data of the endpoint specified with resource_url.
        If either dt_begin or dt_end is None, all available data will be queried.

            resource_url -- url of the endpoint without the base url and without leading or trailing "/"
            earliest_data -- first datapoint in the source collection
            dt_begin -- datetime object for start of data in UTC (no values before: 2011-03-31T22:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        url = f"{self._API_BASE_URL}/data/{resource_url}"
        if (dt_begin is not None) and (dt_end is not None):
            start_of_data = dt_begin.strftime(self._api_date_format)
            end_of_data = dt_end.strftime(self._api_date_format)
            url = f"{self._API_BASE_URL}/data/{resource_url}/{start_of_data}/{end_of_data}"

        response = requests.get(
            url, headers={"Authorization": "Bearer {}".format(self.token)}
        )
        response.raise_for_status()
        df = pd.read_csv(
            io.StringIO(response.text),
            sep=";",
            header=0,
            decimal=",",
            na_values=["N.A."],
        )

        if transform_dates:
            df["von"] = pd.to_datetime(
                df["Datum"] + " " + df["von"] + " " + df["Zeitzone"],
                format="%d.%m.%Y %H:%M %Z",
                utc=True,
            ).dt.tz_localize(None)
            df["bis"] = pd.to_datetime(
                df["Datum"] + " " + df["bis"] + " " + df["Zeitzone"],
                format="%d.%m.%Y %H:%M %Z",
                utc=True,
            ).dt.tz_localize(None)
            df = df.drop(["Datum", "Zeitzone"], axis=1)

        return df

    def redispatch(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /redispatch.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_systemdienstleistungen(
            "redispatch", dt_begin, dt_end, transform_dates
        )

    def kapazitaetsreserve(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /Kapazitaetsreserve.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_systemdienstleistungen(
            "Kapazitaetsreserve", dt_begin, dt_end, transform_dates
        )

    def vorhaltung_krd(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /VorhaltungkRD.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_systemdienstleistungen(
            "VorhaltungkRD", dt_begin, dt_end, transform_dates
        )

    def ausgewiesene_absm(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /AusgewieseneABSM.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_abregelung(
            "AusgewieseneABSM", dt_begin, dt_end, transform_dates
        )

    def zugeteilte_absm(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /ZugeteilteABSM.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_abregelung(
            "ZugeteilteABSM", dt_begin, dt_end, transform_dates
        )

    def erzeugungsverbot(
        self,
        dt_begin: dt.datetime | None = None,
        dt_end: dt.datetime | None = None,
        transform_dates=False,
    ):
        """
        Return a pandas Dataframe with data of the endpoint /Erzeugungsverbot.
        If either dt_begin or dt_end is None, all available data will be queried.

            dt_begin -- datetime object for start of data in UTC (no values before: 2021-10-01T00:00:00)
            dt_end -- datetime object for end of data in UTC
            transform_dates -- The data contains times with date, time and timezone in separate columns
                               if this option resolves to "True" the times will be transformed into two
                               columns "BEGINN" and "ENDE" that contain fully qualified timestamps. (default: False)
        """
        return self._basic_read_abregelung(
            "Erzeugungsverbot", dt_begin, dt_end, transform_dates
        )
